package io.tswf.minecraft.metamorphosis.core.reading.reader

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import groovy.util.logging.Slf4j
import io.tswf.transmutations.expansion.annotations.Expand
import io.tswf.transmutations.staticview.annotation.StaticView
import io.tswf.transmutations.transmute.annotation.Transmute
import io.tswf.transmutations.view.annotation.View
import io.tswf.groovy.bettergroovy.io.tree.IFileTree
import io.tswf.groovy.bettergroovy.io.tree.IFileTreeEntry
import io.tswf.minecraft.metamorphosis.core.model.HooksInfo
import io.tswf.minecraft.metamorphosis.core.reading.FileHooksReadingContext
import io.tswf.minecraft.metamorphosis.core.reading.IFileHooksReadingProvider
import io.tswf.minecraft.metamorphosis.core.util.constants.AnnotationTypes
import io.tswf.minecraft.metamorphosis.core.util.constants.ZipEntryNames
import org.objectweb.asm.ClassReader
import org.objectweb.asm.tree.ClassNode

@Slf4j
@AutoService(IFileHooksReadingProvider)
@CompileStatic
class AsmHooksInfoReader extends AbstractMetaInfInfoReader {

    public static final String CLASS_FILE_POSTFIX = ".class"
    public static final String BASE_PACKAGES_PROPERTY_SEPARATOR = ","
    public static final String BASE_PACKAGES_PROPERTY = "packages"

    AsmHooksInfoReader() {
        super(ZipEntryNames.HOOKS_AUTO)
    }

    @Override
    void readMetaInfEntry(FileHooksReadingContext context, IFileTree fileTree, IFileTreeEntry fileTreeEntry) {
        fileTreeEntry.inputStream.newReader("UTF-8").doThenClose { zipEntryReader ->
            def properties = new Properties()
            properties.load(zipEntryReader)

            def packages = properties.getJoinedProperty(BASE_PACKAGES_PROPERTY, BASE_PACKAGES_PROPERTY_SEPARATOR) ?: []
            def (serverHooks, clientHooks) = readClassesInPackages(packages.toSet(), fileTree, context.hooksInfo.file)

            context.hooksInfo?.let {
                it.server.addAll(serverHooks)
                it.client.addAll(clientHooks)
            }
        }
    }

    Tuple2<HooksInfo, HooksInfo> readClassesInPackages(Set<String> packages, IFileTree fileTree, File file) {
        log.info("Scanning for hooks file '${file.name}' in ${packages.isEmpty() ? "all" : packages} packages...")

        def clientHooks = new HooksInfo()
        def serverHooks = new HooksInfo()

        def packageNamePrefixes = packages.map { it.replace(".", "/").concat("/") }

        def classZipEntries = fileTree.entries
                .with { stream ->
                    packageNamePrefixes.isEmpty() ? stream :
                            stream.filter { treeEntry ->
                                packageNamePrefixes.any {packageNamePrefix -> treeEntry.name.startsWith(packageNamePrefix) && treeEntry.name != packageNamePrefix }
                            }
                }
                .filter { it.name.endsWith(CLASS_FILE_POSTFIX) }
                .toSet()

        for (def classZipEntry : classZipEntries) {
            try {
                classZipEntry.inputStream.doThenClose { classInputStream ->
                    def classNode = new ClassNode()
                        .tap {
                            new ClassReader(classInputStream).accept(it, ClassReader.SKIP_FRAMES)
                        }

                    def isClientClass = classNode.hasAnnotation(AnnotationTypes.CLIENT_ONLY_ANNOTATION)
                    def hooksIntoAdd = isClientClass ? clientHooks : serverHooks
                    def classType = (String) null

                    def classesIntoAdd = (List<String>) null

                    if (classNode.hasAnnotation(View)) {
                        classType = "view"
                        classesIntoAdd = hooksIntoAdd.views
                    } else if (classNode.hasAnnotation(StaticView)) {
                        classType = "staticView"
                        classesIntoAdd = hooksIntoAdd.staticViews
                    } else if (classNode.hasAnnotation(Expand)) {
                        classType = "expansion"
                        classesIntoAdd = hooksIntoAdd.expansions
                    } else if (classNode.methods.any { it.hasAnnotation(Transmute) }) {
                        classType = "transmutation"
                        classesIntoAdd = hooksIntoAdd.transmutations
                    }

                    if (classesIntoAdd != null) {
                        log.info("Found ${isClientClass ? "client" : "server"} '${classType}' class '${classNode.javaName}' in '${file.absoluteCanonicalFile.path}'")
                        classesIntoAdd << classNode.javaName
                    }
                }
            } catch (Throwable e) {
                log.warn("Exception occurred while trying to read class candidate zip entry named '${classZipEntry.name}' from file '${file.absoluteCanonicalFile.path}'", e)
            }
        }

        return new Tuple2<HooksInfo, HooksInfo>(serverHooks, clientHooks)
    }
}
