package io.tswf.minecraft.metamorphosis.core.reading.reader

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import groovy.util.logging.Slf4j
import io.tswf.groovy.bettergroovy.io.tree.IFileTree
import io.tswf.groovy.bettergroovy.io.tree.IFileTreeEntry
import io.tswf.minecraft.metamorphosis.core.model.HooksInfo
import io.tswf.minecraft.metamorphosis.core.reading.FileHooksReadingContext
import io.tswf.minecraft.metamorphosis.core.reading.IFileHooksReadingProvider
import io.tswf.minecraft.metamorphosis.core.util.constants.ZipEntryNames

@Slf4j
@CompileStatic
@AutoService(IFileHooksReadingProvider)
class PropertiesHooksInfoReader extends AbstractMetaInfInfoReader {

    public static final String CLIENT_PREFIX = "client"
    public static final String SERVER_PREFIX = "server"

    public static final String STATIC_VIEWS_PROPERTY = "static-views"
    public static final String VIEWS_PROPERTY = "views"
    public static final String EXPANSIONS_PROPERTY = "expansions"
    public static final String TRANSMUTATIONS_PROPERTY = "transmutations"

    public static final String CLASS_NAMES_SEPARATOR = ","

    PropertiesHooksInfoReader() {
        super(ZipEntryNames.HOOKS_PROPERTIES)
    }

    @Override
    void readMetaInfEntry(FileHooksReadingContext context, IFileTree fileTree, IFileTreeEntry fileTreeEntry) {
        fileTreeEntry.getInputStream().newReader("UTF-8").doThenClose { zipEntryReader ->
            def properties = new Properties()
            properties.load(zipEntryReader)

            def clientHooks = readSideHooks(CLIENT_PREFIX, properties)
            def serverHooks = readSideHooks(SERVER_PREFIX, properties)

            context.hooksInfo.let {
                it.client.addAll(clientHooks)
                it.server.addAll(serverHooks)
            }
        }
    }

    static HooksInfo readSideHooks(String sidePrefix, Properties properties) {
        def result = new HooksInfo()

        properties.getProperty(getPropertyName(sidePrefix), VIEWS_PROPERTY)?.let {
            result.views.addAll(readJoinedClassNames(it))
        }

        properties.getProperty(getPropertyName(sidePrefix), STATIC_VIEWS_PROPERTY)?.let {
            result.views.addAll(readJoinedClassNames(it))
        }

        properties.getProperty(getPropertyName(sidePrefix), EXPANSIONS_PROPERTY)?.let {
            result.expansions.addAll(readJoinedClassNames(it))
        }

        properties.getProperty(getPropertyName(sidePrefix), TRANSMUTATIONS_PROPERTY)?.let {
            result.transmutations.addAll(readJoinedClassNames(it))
        }

        return result
    }

    static List<String> readJoinedClassNames(String classNamesString) {
        return classNamesString.split(CLASS_NAMES_SEPARATOR)
            .stream()
                .map { it.stripAll() }
                .distinct()
                .toList()
    }

    static String getPropertyName(String... path) {
        return path.join(".")
    }
}
