package io.tswf.minecraft.metamorphosis.core.remap

import com.google.auto.service.AutoService
import groovy.transform.CompileStatic
import groovy.transform.Memoized
import io.tswf.minecraft.metamorphosis.core.exception.TinyRemapperAdapterConfigurationListenerException
import io.tswf.transmutations.common.remapping.ITransmutationRemappingStrategy
import net.fabricmc.mappingio.MappingReader
import net.fabricmc.mappingio.format.MappingFormat
import net.fabricmc.mappingio.tree.MemoryMappingTree
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method

@CompileStatic
class TinyRemappingAdapter {

    @Lazy
    static List<ITinyRemapperConfigurationListener> configurationListeners = ServiceLoader.load(ITinyRemapperConfigurationListener)
        .sortBy(true) { it.priority }
        .toList()

    static boolean isRemappingEnabled = [
            ClassLoader.getSystemClassLoader(),
            ClassLoader.contextClassloader()
    ].any {
        it.getResource("aaa.class") != null
    }

    static MemoryMappingTree theRemapper
    static int namespaceFrom = -1
    static int namespaceTo = -1

    @Memoized
    static String _ensureConfigured() {
        for (def listener : configurationListeners) {
            try {
                listener.onTinyRemapperConfigurationStarts()
            } catch (Throwable e) {
                throw new TinyRemapperAdapterConfigurationListenerException(e, listener)
            }
        }

        if (theRemapper == null) {
            throw new IllegalStateException("Remapper was not configured by any listener in ${configurationListeners}!")
        }

        return null
    }

    static Type mapType(Type type) {
        _ensureConfigured()
        if (!isRemappingEnabled) {
            return type
        }

        def mappedInternalName = theRemapper.mapClassName(type.internalName, namespaceFrom, namespaceTo)
        if (mappedInternalName == null || mappedInternalName == type.internalName) {
            return type
        }

        return Type.getObjectType(mappedInternalName)
    }

    static Type mapTypeReverse(Type type) {
        _ensureConfigured()
        if (!isRemappingEnabled) {
            return type
        }

        def mappedInternalName = theRemapper.mapClassName(type.internalName, namespaceTo, namespaceFrom)
        if (mappedInternalName == null || mappedInternalName == type.internalName) {
            return type
        }

        return Type.getObjectType(mappedInternalName)
    }

    static Method mapMethod(Type rawOwner, Method rawMethod) {
        _ensureConfigured()
        if (!isRemappingEnabled) {
            return rawMethod
        }

        def reversedOwner = mapTypeReverse(rawOwner)
        def reversedMethod = new StringBuilder().tap {
            def methodReturnType = rawMethod.returnType
            def methodParameterTypes = rawMethod.argumentTypes

            append("(")
            for (def methodParameterType : methodParameterTypes) {
                append(mapTypeReverse(methodParameterType).descriptor)
            }
            append(")")

            append(mapTypeReverse(methodReturnType).descriptor)
        }
                .with {
                    new Method(rawMethod.name, it.toString())
                }

        def mappedMethod = theRemapper.getMethod(reversedOwner.internalName, reversedMethod.name, reversedMethod.descriptor, namespaceFrom)
        if (mappedMethod != null) {
            def mappedMethodName = mappedMethod.getName(namespaceTo)
            def mappedMethodDesc = mappedMethod.getDesc(namespaceTo)

            if (mappedMethodDesc != rawMethod.descriptor || mappedMethodName != rawMethod.name) {
                return new Method(mappedMethodName, mappedMethodDesc)
            }
        }
        return rawMethod
    }

    static Tuple2<String, Type> mapField(Type fieldOwner, String fieldName, Type rawFieldType) {
        _ensureConfigured()
        if (isRemappingEnabled) {
            fieldOwner = mapTypeReverse(fieldOwner)
            rawFieldType = mapTypeReverse(rawFieldType)

            def mappedField = theRemapper.getField(fieldOwner.internalName, fieldName, rawFieldType.descriptor, namespaceFrom)

            if (mappedField != null) {
                def mappedFieldName = mappedField.getName(namespaceTo)
                def mappedFieldDesc = mappedField.getDesc(namespaceTo)

                if (mappedFieldName != fieldName || mappedFieldDesc != rawFieldType.descriptor) {
                    return new Tuple2<String, Type>(mappedFieldName, Type.getType(mappedFieldDesc))
                }
            }
        }

        return new Tuple2<String, Type>(fieldName, rawFieldType)
    }

    static void initializeRemapper(String mappingsLocation, String nameFrom, String nameTo, MappingFormat mappingFormat = MappingFormat.TINY_2) {
        def tree = new MemoryMappingTree()
        def mappingsInputStream = ClassLoader.contextClassloader().getResourceAsStream(mappingsLocation)

        if (mappingsInputStream == null) {
            throw new IOException("There is no mappings located at '${mappingsLocation}'!")
        }

        MappingReader.read(mappingsInputStream.newReader(), mappingFormat, tree)

        theRemapper = tree
        namespaceFrom = theRemapper.dstNamespaces.indexOf(nameFrom)
        namespaceTo = theRemapper.dstNamespaces.indexOf(nameTo)
    }

    @AutoService(ITransmutationRemappingStrategy)
    static class TinyToSrgViewsRemapperAdapter implements ITransmutationRemappingStrategy {

        @Override
        Type mapType(Type type) {
            return TinyRemappingAdapter.mapType(type)
        }

        @Override
        Method mapMethod(Type type, Method method) {
            return TinyRemappingAdapter.mapMethod(type, method)
        }

        @Override
        Tuple2<String, Type> mapField(Type fieldOwner, String fieldName, Type fieldType) {
            return TinyRemappingAdapter.mapField(fieldOwner, fieldName, fieldType)
        }

        @Override
        String getName() {
            return this.class.name
        }

        @Override
        int getPriority() {
            return 0
        }
    }
}

@CompileStatic
interface ITinyRemapperConfigurationListener {
    void onTinyRemapperConfigurationStarts()
    int getPriority()
}